import os
import sys

from mykit.kit.utils import printer


def inspect(CONTAINER_DIR_PTH, WORKFLOW_FILE_PTH):
    
    ## the container ('docs/' folder) should not exist
    if os.path.isdir(CONTAINER_DIR_PTH):
        printer(f'ERROR: Folder {repr(CONTAINER_DIR_PTH)} already exists.')
        sys.exit(1)

    ## the GitHub Action workflow file 'rebuild-docs.yml' should not exist
    if os.path.isfile(WORKFLOW_FILE_PTH):
        printer(f'ERROR: File {repr(WORKFLOW_FILE_PTH)} already exists.')
        sys.exit(1)


import os

from mykit.kit.keycrate import KeyCrate
from mykit.kit.utils import printer

from ..constants import __version__, SOFTWARE_DIST_NAME, SETTINGS_KEYS
from .inspect import inspect_the_container
from .update_header import update_header
from .update_footer import update_footer
from .update_jekyll_config import update_jekyll_config
from .update_404_and_favicon import update_404_and_favicon
from .copying_template import copying_template
from .update_sass_constants import update_sass_constants
from .update_docs_tree import update_docs_tree


def run(container, dock):
    ## `container`: the 'docs/' folder (in main branch)
    ## `dock`: the 'docs' branch
    printer(f"INFO: Running '_compile' command ({SOFTWARE_DIST_NAME}-v{__version__})")

    C_TREE = os.path.join(container, 'tree')
    C_404 = os.path.join(container, '404.md')
    C_ICON = os.path.join(container, 'favicon.png')
    C_SETTINGS = os.path.join(container, 'settings.txt')

    D__INCLUDES = os.path.join(dock, '_includes')
    D__PAGES = os.path.join(dock, '_pages')
    D_HEADER = os.path.join(dock, '_includes', 'header.html')
    D_FOOTER = os.path.join(dock, '_includes', 'footer.html')
    D_JEKYLL_CONFIG = os.path.join(dock, '_config.yml')
    D_404 = os.path.join(dock, '404.md')
    D_ICON = os.path.join(dock, 'favicon.png')
    D_SASS_CONSTANTS = os.path.join(dock, '_sass', 'constants.scss')

    ## inspection
    inspect_the_container(container)

    ## parse the settings
    cfg = KeyCrate(C_SETTINGS, True, True, SETTINGS_KEYS, SETTINGS_KEYS)


    if not os.path.isdir(D__INCLUDES):  # handle init case: initially, '_includes' folder doesn't exist in docs branch
        printer(f'DEBUG: Creating dir {repr(D__INCLUDES)}.')
        os.mkdir(D__INCLUDES)
    update_header(C_TREE, D_HEADER)
    update_footer(D_FOOTER, cfg.show_credit)

    update_jekyll_config(D_JEKYLL_CONFIG, cfg.author, cfg._gh_username, cfg._gh_repo)
    update_404_and_favicon(C_404, C_ICON, D_404, D_ICON)

    copying_template(dock)
    update_sass_constants(D_SASS_CONSTANTS, cfg.color_hue)

    update_docs_tree(C_TREE, D__PAGES, cfg.lowercase_the_url)


import argparse

from .constants import __version__, SOFTWARE_DIST_NAME
from .wizard import run as run_init
from .upgrade import run as run_upgrade
from .compile import run as run_compile


def main():

    psr = argparse.ArgumentParser(
        prog=SOFTWARE_DIST_NAME,
        usage=(
            '\n'
            '├─ Run `%(prog)s init`   : Set up NICS environment\n'
            '└─ Run `%(prog)s upgrade`: Reconfigure NICS environment'
        ),
        formatter_class=argparse.RawTextHelpFormatter  # to use line breaks (\n) in the help message
    )
    psr.add_argument(
        '-v', '--version', action='version', version=f'%(prog)s-{__version__}',
        help='show software version'
    )
    subpsr = psr.add_subparsers(dest='cmd', help=argparse.SUPPRESS)  # `help=argparse.SUPPRESS` to hide the help message

    ## command 'init'
    subpsr.add_parser('init', help=argparse.SUPPRESS)

    ## command 'upgrade'
    subpsr.add_parser('upgrade', help=argparse.SUPPRESS)

    ## command '_compile' (that users shouldn't run)
    c = subpsr.add_parser('_compile', help=argparse.SUPPRESS)
    c.add_argument('container')
    c.add_argument('dock')

    args = psr.parse_args()

    if args.cmd == 'init':
        run_init()
    elif args.cmd == 'upgrade':
        run_upgrade()
    elif args.cmd == '_compile':
        run_compile(args.container, args.dock)


import os
import re
import shutil

from mykit.kit.utils import printer


def update_recursively(D__PAGES, lowercase_the_url, pth, base):
    printer(f'DEBUG: Updating docs-tree, pth: {repr(pth)}, base: {repr(base)}')

    for i in os.listdir(pth):
        pth2 = os.path.join(pth, i)

        ## handle non-markdown files
        if os.path.isfile(pth2) and (not i.endswith('.md')):
            dst = os.path.join(D__PAGES, os.sep.join(filter(lambda s:s!='', base.split('/'))), i)
            printer(f'DEBUG: Copying non-markdown files from {repr(pth2)} to {repr(dst)}.')
            shutil.copy(pth2, dst)
            continue

        ## <handling the index.md>
        if i == 'index.md':
            if base == '/':  # homepage
                dst = os.path.join(D__PAGES, 'index.md')
                text = (
                    '---\n'
                    'permalink: /\n'
                    'layout: main\n'
                    'title: Home\n'
                    '---\n\n'
                )
            else:
                dst = os.path.join(D__PAGES, os.sep.join(filter(lambda s:s!='', base.split('/'))), 'index.md')
                text = (
                    '---\n'
                    f'permalink: {base}\n'
                    'layout: main\n'
                    f"title: {list(filter(lambda s:s!='', base.split('/')))[-1]}\n"
                    '---\n\n'
                )
            with open(pth2, 'r') as f: text += f.read()
            with open(dst, 'w') as f: f.write(text)
            printer(f'DEBUG: Updated index.md from {repr(pth2)} to {repr(dst)}.')
            continue
        ## </handling the index.md>

        res = re.match(r"^(?:\d+ - )?(?P<name>[\w \-'&\(\).]+)(?:\.md)?$", i)
        if res is None: raise AssertionError(f'Invalid docs-tree format: {repr(i)}')
        name = res.group('name')
        url = name.replace(' ', '-').replace("'", '').replace('&', 'and').replace('(', '').replace(')', '').replace('.', '-')
        if lowercase_the_url: url = url.lower()
        printer(f'DEBUG: name: {repr(name)}; url: {repr(url)}')

        if os.path.isdir(pth2):
            ## if the folder doesn't exist, create a new one.
            dir = os.path.join(D__PAGES, os.sep.join(filter(lambda s:s!='', base.split('/'))), url)
            if not os.path.isdir(dir):
                os.mkdir(dir)
                printer(f'DEBUG: Dir created: {repr(dir)}.')
            ## do it again
            update_recursively(D__PAGES, lowercase_the_url, pth2, base+url+'/')
        else:
            dst = os.path.join(D__PAGES, os.sep.join(filter(lambda s:s!='', base.split('/'))), f'{name}.md')
            text = (
                '---\n'
                f'permalink: {base}{url}/\n'
                'layout: main\n'
                f'title: {name}\n'
                '---\n\n'
            )
            with open(pth2, 'r') as f: text += f.read()
            with open(dst, 'w') as f: f.write(text)
            printer(f'DEBUG: Updated docs-tree file from {repr(pth2)} to {repr(dst)}.')


def update_docs_tree(C_TREE, D__PAGES, lowercase_the_url):

    ## delete the old '_pages' folder in docs branch
    if os.path.isdir(D__PAGES):  # reminder: initially, '_pages' doesn't exist
        printer(f'DEBUG: Deleting {repr(D__PAGES)} recursively.')
        shutil.rmtree(D__PAGES)

    printer(f'DEBUG: Creating dir {repr(D__PAGES)}.')
    os.mkdir(D__PAGES)

    update_recursively(D__PAGES, lowercase_the_url, C_TREE, '/')


import os


def inspect_the_container(container):

    ## homepage 'tree/index.md' file must exist
    if not os.path.isfile( os.path.join(container, 'tree', 'index.md') ):
        raise AssertionError("Couldn't find 'tree/index.md' in the container.")

    ## settings file must exist
    if not os.path.isfile( os.path.join(container, 'settings.txt') ):
        raise AssertionError("Couldn't find 'settings.txt' in the container.")


import os
import sys

from mykit.kit.keycrate import KeyCrate
from mykit.kit.utils import printer

from ..constants import __version__, SETTINGS_KEYS
from ..wizard.workflow_writer import _writer as the_workflow_writer
from ..wizard.settings_writer import full_writer as the_settings_writer


def run():

    CWD = os.getcwd()

    WORKFLOW_FILE_PTH = os.path.join(CWD, '.github', 'workflows', 'rebuild-docs.yml')
    SETTINGS_FILE_PTH = os.path.join(CWD, 'docs', 'settings.txt')

    if not os.path.isfile(WORKFLOW_FILE_PTH):
        printer(f'ERROR: Workflow file not found: {repr(WORKFLOW_FILE_PTH)}')
        sys.exit(1)

    if not os.path.isfile(SETTINGS_FILE_PTH):
        printer(f'ERROR: Settings file not found: {repr(SETTINGS_FILE_PTH)}')
        sys.exit(1)

    ## parse the settings
    cfg = KeyCrate(SETTINGS_FILE_PTH, True, True, SETTINGS_KEYS, SETTINGS_KEYS)

    if __version__ == cfg._nics_version:
        printer(f'INFO: Everything is up to date.')
        sys.exit(0)

    if __version__.split('.')[0] != cfg._nics_version.split('.')[0]:
        printer(f"ERROR: Cannot upgrade across major versions. Please run 'nics init' instead.")
        sys.exit(1)


    text = the_workflow_writer(cfg.author, cfg._email, cfg._gh_repo, cfg._main_branch_name)
    with open(WORKFLOW_FILE_PTH, 'w') as f: f.write(text)
    printer(f'INFO: Workflow file {repr(WORKFLOW_FILE_PTH)} is updated.')

    text = the_settings_writer(
        cfg.author, cfg.color_hue, cfg.lowercase_the_url, cfg.show_credit,
        cfg._email, cfg._gh_username, cfg._gh_repo, cfg._main_branch_name
    )
    with open(SETTINGS_FILE_PTH, 'w') as f: f.write(text)
    printer(f'INFO: Settings file {repr(SETTINGS_FILE_PTH)} is updated.')
    
    printer(f'INFO: Upgrade finished.')


import os
import shutil

from mykit.kit.utils import printer


def update_404_and_favicon(C_404, C_ICON, D_404, D_ICON):

    ## 404 page
    printer(f"DEBUG: Attempting to copy '404.md'.")
    if os.path.isfile(C_404):
        printer(f'DEBUG: Rewriting 404.md from {repr(C_404)} to {repr(D_404)}.')
        text = (
            '---\n'
            'permalink: /404.html\n'
            'layout: main\n'
            'title: Page not found\n'
            '---\n\n'
        )
        with open(C_404, 'r') as f: text += f.read()
        with open(D_404, 'w') as f: f.write(text)
        printer(f'INFO: 404.md is rewritten.')

    ## favicon
    printer(f"DEBUG: Attempting to copy 'favicon.png'.")
    if os.path.isfile(C_ICON):
        printer(f'DEBUG: Copying favicon from {repr(C_ICON)} to {repr(D_ICON)}.')
        shutil.copy(C_ICON, D_ICON)
        printer(f'INFO: Favicon is renewed.')


import os
import shutil

from mykit.kit.utils import printer

from ..constants import TMPL_WEB_DIR_PTH


def copying_template(dock):

    LAYOUTS = os.path.join(dock, '_layouts')
    SASS = os.path.join(dock, '_sass')
    SCRIPTS = os.path.join(dock, 'scripts')

    ## handle the case when the template already exists
    if os.path.isdir(LAYOUTS): shutil.rmtree(LAYOUTS)
    if os.path.isdir(SASS): shutil.rmtree(SASS)
    if os.path.isdir(SCRIPTS): shutil.rmtree(SCRIPTS)

    shutil.copytree(
        os.path.join(TMPL_WEB_DIR_PTH, '_layouts'),
        LAYOUTS
    )
    shutil.copytree(
        os.path.join(TMPL_WEB_DIR_PTH, '_sass'),
        SASS
    )
    shutil.copytree(
        os.path.join(TMPL_WEB_DIR_PTH, 'scripts'),
        SCRIPTS
    )
    printer(f'INFO: Template copied.')


from mykit.kit.utils import printer


def _writer(author, gh_username, gh_repo):
	return f"""

# website

baseurl: /{gh_repo}
url: https://{gh_username}.github.io/{gh_repo}


# personal

title: {gh_repo}
author:
  name: {author}


# internal

include: [_pages, _sass, scripts]

permalink: pretty

sass:
  style: compact # possible values: nested expanded compact compressed
  sass_dir: _sass
"""


def update_jekyll_config(D_JEKYLL_CONFIG, author, gh_username, gh_repo):

	text = _writer(author, gh_username, gh_repo)
	with open(D_JEKYLL_CONFIG, 'w') as f: f.write(text)
	printer(f"INFO: Updated Jekyll '_config.yml' file {repr(D_JEKYLL_CONFIG)}.")


import random

from mykit.kit.utils import printer

from ..constants import __version__


def full_writer(author, color_hue, lowercase_the_url, show_credit, email, gh_username, gh_repo, main_branch_name):
    return f"""
#-- Welcome to NICS settings!
#----------------------------

#-- Everything starts with "#--" is a comment.
#-- Read documentation at https://nvfp.github.io/now-i-can-sleep


author: '{author}'
color_hue: {color_hue}
lowercase_the_url: {lowercase_the_url}
show_credit: {show_credit}


#-- The below variables are for NICS internal use only and should not be changed.

_email: '{email}'
_gh_username: '{gh_username}'
_gh_repo: '{gh_repo}'
_main_branch_name: '{main_branch_name}'

_nics_version: '{__version__}'
"""


def _writer(author, email, gh_username, gh_repo, main_branch_name):
    return f"""
#-- Welcome to NICS settings!
#----------------------------

#-- Everything starts with "#--" is a comment.
#-- Read documentation at https://nvfp.github.io/now-i-can-sleep


author: '{author}'
color_hue: {random.randint(0, 359)}
lowercase_the_url: True
show_credit: True


#-- The below variables are for NICS internal use only and should not be changed.

_email: '{email}'
_gh_username: '{gh_username}'
_gh_repo: '{gh_repo}'
_main_branch_name: '{main_branch_name}'

_nics_version: '{__version__}'
"""


def settings_writer(pth, author, email, gh_username, gh_repo, main_branch_name):
    printer(f'INFO: Writing settings file.')

    text = _writer(author, email, gh_username, gh_repo, main_branch_name)
    with open(pth, 'w') as f:
        f.write(text)

    printer(f'INFO: Done, {repr(pth)} is created.')